from os import PathLike
from unittest.mock import patch

import pytest
from infoblox_client.objects import HostRecord

from gso.products import ProductName
from gso.products.product_blocks.iptrunk import IptrunkType
from gso.products.product_types.iptrunk import Iptrunk
from gso.services.subscriptions import get_product_id_by_name
from gso.utils.shared_enums import Vendor
from gso.utils.types.interfaces import PhysicalPortCapacity
from test import USER_CONFIRM_EMPTY_FORM
from test.services.conftest import MockedNetboxClient, MockedSharePointClient
from test.workflows import (
    assert_complete,
    assert_failed,
    assert_lso_interaction_failure,
    assert_lso_interaction_success,
    assert_suspended,
    extract_state,
    resume_workflow,
    run_workflow,
)


@pytest.fixture()
def _netbox_client_mock():
    # Mock NetboxClient methods
    with (
        patch("gso.services.netbox_client.NetboxClient.get_device_by_name") as mock_get_device_by_name,
        patch("gso.services.netbox_client.NetboxClient.get_available_interfaces") as mock_get_available_interfaces,
        patch("gso.services.netbox_client.NetboxClient.get_available_lags") as mock_get_available_lags,
        patch("gso.services.netbox_client.NetboxClient.create_interface") as mock_create_interface,
        patch("gso.services.netbox_client.NetboxClient.attach_interface_to_lag") as mock_attach_interface_to_lag,
        patch("gso.services.netbox_client.NetboxClient.reserve_interface") as mock_reserve_interface,
        patch("gso.services.netbox_client.NetboxClient.allocate_interface") as mock_allocate_interface,
    ):
        mock_get_device_by_name.return_value = MockedNetboxClient().get_device_by_name()
        mock_get_available_interfaces.return_value = MockedNetboxClient().get_available_interfaces()
        mock_get_available_lags.return_value = MockedNetboxClient().get_available_lags()
        mock_create_interface.return_value = MockedNetboxClient().create_interface()
        mock_attach_interface_to_lag.return_value = MockedNetboxClient().attach_interface_to_lag()
        mock_reserve_interface.return_value = MockedNetboxClient().reserve_interface()
        mock_allocate_interface.return_value = MockedNetboxClient().allocate_interface()

        yield


@pytest.fixture()
def input_form_wizard_data(request, router_subscription_factory, faker):
    vendor = getattr(request, "param", Vendor.NOKIA)
    router_side_a = router_subscription_factory()

    # Set side b router to Juniper
    if vendor == Vendor.JUNIPER:
        router_side_b = router_subscription_factory(vendor=Vendor.JUNIPER)
        side_b_members = faker.link_members_juniper()[0:2]
    else:
        router_side_b = router_subscription_factory()
        side_b_members = [
            {"interface_name": f"Interface{interface}", "interface_description": faker.sentence()}
            for interface in range(2)
        ]

    create_ip_trunk_step = {
        "tt_number": faker.tt_number(),
        "geant_s_sid": faker.geant_sid(),
        "iptrunk_type": IptrunkType.DARK_FIBER,
        "iptrunk_description": faker.sentence(),
        "iptrunk_speed": PhysicalPortCapacity.HUNDRED_GIGABIT_PER_SECOND,
        "iptrunk_number_of_members": 2,
    }
    create_ip_trunk_confirm_step = {"iptrunk_minimum_links": 1}
    create_ip_trunk_side_a_router_name = {"side_a_node_id": router_side_a}
    create_ip_trunk_side_a_step = {
        "side_a_ae_iface": "lag-1",
        "side_a_ae_geant_a_sid": None,
        "side_a_ae_members": [
            {
                "interface_name": f"Interface{interface}",
                "interface_description": faker.sentence(),
            }
            for interface in range(2)
        ],
    }
    create_ip_trunk_side_b_router_name = {"side_b_node_id": router_side_b}
    create_ip_trunk_side_b_step = {
        "side_b_ae_iface": "lag-4",
        "side_b_ae_geant_a_sid": faker.geant_sid(),
        "side_b_ae_members": side_b_members,
    }
    summary_view_step = {}

    return [
        create_ip_trunk_step,
        create_ip_trunk_confirm_step,
        create_ip_trunk_side_a_router_name,
        create_ip_trunk_side_a_step,
        create_ip_trunk_side_b_router_name,
        create_ip_trunk_side_b_step,
        summary_view_step,
    ]


@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.create_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.find_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.ping")
@patch("gso.workflows.iptrunk.create_iptrunk.SharePointClient")
def test_successful_iptrunk_creation_with_standard_lso_result(
    mock_sharepoint_client,
    mock_ping,
    mock_find_host_by_ip,
    mock_create_host,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_execute_playbook,
    responses,
    input_form_wizard_data,
    faker,
    _netbox_client_mock,  # noqa: PT019
    data_config_filename: PathLike,
    test_client,
):
    mock_create_host.return_value = None
    mock_allocate_v4_network.return_value = faker.ipv4_network(min_subnet=31, max_subnet=31)
    mock_allocate_v6_network.return_value = faker.ipv6_network(min_subnet=126, max_subnet=126)
    mock_find_host_by_ip.return_value = None
    mock_ping.return_value = False
    mock_sharepoint_client.return_value = MockedSharePointClient

    product_id = get_product_id_by_name(ProductName.IP_TRUNK)
    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, process_stat, step_log = run_workflow("create_iptrunk", initial_site_data)

    for _ in range(6):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    assert_complete(result)

    state = extract_state(result)
    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    sorted_sides = sorted([
        subscription.iptrunk.iptrunk_sides[0].iptrunk_side_node.router_site.site_name,
        subscription.iptrunk.iptrunk_sides[1].iptrunk_side_node.router_site.site_name,
    ])
    assert subscription.status == "provisioning"
    assert subscription.description == (
        f"IP trunk {sorted_sides[0]} {sorted_sides[1]}, geant_s_sid:{input_form_wizard_data[0]["geant_s_sid"]}"
    )

    assert mock_execute_playbook.call_count == 6
    #  We search for 6 hosts in total, 2 in a /31 and 4 in a /126
    assert mock_find_host_by_ip.call_count == 6
    assert mock_ping.call_count == 6


@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.find_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.ping")
def test_iptrunk_creation_fails_when_lso_return_code_is_one(
    mock_ping,
    mock_find_host_by_ip,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_execute_playbook,
    responses,
    input_form_wizard_data,
    faker,
    _netbox_client_mock,  # noqa: PT019
    data_config_filename: PathLike,
):
    mock_allocate_v4_network.return_value = faker.ipv4_network(min_subnet=31, max_subnet=31)
    mock_allocate_v6_network.return_value = faker.ipv6_network(min_subnet=126, max_subnet=126)
    mock_find_host_by_ip.return_value = None
    mock_ping.return_value = False
    product_id = get_product_id_by_name(ProductName.IP_TRUNK)

    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, process_stat, step_log = run_workflow("create_iptrunk", initial_site_data)

    result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_lso_interaction_failure(result, process_stat, step_log)

    assert mock_execute_playbook.call_count == 2
    assert mock_find_host_by_ip.call_count == 6
    assert mock_ping.call_count == 6


@pytest.mark.parametrize("input_form_wizard_data", [Vendor.JUNIPER], indirect=True)
@pytest.mark.workflow()
@patch("gso.services.lso_client._send_request")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.create_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.find_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.ping")
@patch("gso.workflows.iptrunk.create_iptrunk.SharePointClient")
def test_successful_iptrunk_creation_with_juniper_interface_names(
    mock_sharepoint_client,
    mock_ping,
    mock_find_host_by_ip,
    mock_create_host,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    mock_execute_playbook,
    responses,
    input_form_wizard_data,
    faker,
    data_config_filename: PathLike,
    _netbox_client_mock,  # noqa: PT019
    test_client,
):
    mock_create_host.return_value = None
    mock_allocate_v4_network.return_value = faker.ipv4_network(min_subnet=31, max_subnet=31)
    mock_allocate_v6_network.return_value = faker.ipv6_network(min_subnet=126, max_subnet=126)
    mock_find_host_by_ip.return_value = None
    mock_ping.return_value = False

    mock_sharepoint_client.return_value = MockedSharePointClient
    product_id = get_product_id_by_name(ProductName.IP_TRUNK)
    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, process_stat, step_log = run_workflow("create_iptrunk", initial_site_data)

    for _ in range(6):
        result, step_log = assert_lso_interaction_success(result, process_stat, step_log)

    assert_suspended(result)
    result, step_log = resume_workflow(process_stat, step_log, input_data=USER_CONFIRM_EMPTY_FORM)

    assert_complete(result)
    assert mock_execute_playbook.call_count == 6
    assert mock_find_host_by_ip.call_count == 6
    assert mock_ping.call_count == 6


@pytest.mark.workflow()
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.find_host_by_ip")
def test_iptrunk_creation_with_taken_dns_record(
    mock_find_host_by_ip,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    input_form_wizard_data,
    faker,
    _netbox_client_mock,  # noqa: PT019
):
    mock_allocate_v4_network.return_value = faker.ipv4_network(min_subnet=31, max_subnet=31)
    mock_allocate_v6_network.return_value = faker.ipv6_network(min_subnet=126, max_subnet=126)
    mock_find_host_by_ip.return_value = HostRecord(connector=None, hostname="fake.internal")

    product_id = get_product_id_by_name(ProductName.IP_TRUNK)
    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, _, _ = run_workflow("create_iptrunk", initial_site_data)

    assert_failed(result)

    state = extract_state(result)
    assert (
        state["error"] == "One or more hosts in the assigned IPv4 network are already registered, please investigate."
    )

    #  We search for 2 hosts in a /31 and then fail the workflow
    assert mock_find_host_by_ip.call_count == 2


@pytest.mark.workflow()
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v6_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.allocate_v4_network")
@patch("gso.workflows.iptrunk.create_iptrunk.infoblox.find_host_by_ip")
@patch("gso.workflows.iptrunk.create_iptrunk.ping")
def test_iptrunk_creation_with_taken_ip_address(
    mock_ping,
    mock_find_host_by_ip,
    mock_allocate_v4_network,
    mock_allocate_v6_network,
    input_form_wizard_data,
    faker,
    _netbox_client_mock,  # noqa: PT019
):
    mock_allocate_v4_network.return_value = faker.ipv4_network(min_subnet=31, max_subnet=31)
    mock_allocate_v6_network.return_value = faker.ipv6_network(min_subnet=126, max_subnet=126)
    mock_find_host_by_ip.return_value = None
    mock_ping.return_value = True

    product_id = get_product_id_by_name(ProductName.IP_TRUNK)
    initial_site_data = [{"product": product_id}, *input_form_wizard_data]
    result, _, _ = run_workflow("create_iptrunk", initial_site_data)

    assert_failed(result)

    state = extract_state(result)
    assert (
        state["error"] == "One or more hosts in the assigned IPv4 network are responding to ping, please investigate."
    )

    assert mock_find_host_by_ip.call_count == 6
    #  We ping 2 hosts in a /31 and then fail the workflow
    assert mock_ping.call_count == 2
